module Deus.GLUtils where

import Control.Monad
import Control.Monad.Trans

import Data.Map(Map)
import qualified Data.Map as M

import Graphics.Rendering.OpenGL

import Deus.With
import Deus.Lens

-- some generic GL Utils

-- set up a rather standard 3D project matrix
reshape screenSize@(Size w h) = do
  viewport $= ((Position 0 0), screenSize)
  matrixMode $= Projection
  loadIdentity
  let near   = 1
      far    = 10
      fov    = 40
      ang    = (fov*pi)/(360)
      top    = near / ( cos(ang) / sin(ang) )
      aspect = fromIntegral(w)/fromIntegral(h)
      right = top*aspect
  frustum (-right) right (-top) top near far
  matrixMode $= Modelview 0

-- set up 2D projection matrix based on screen pixels
with2DByPixel a = 
  preservingMatrix$do
    loadIdentity
    matrixMode $= Projection
    r <- preservingMatrix$do 
      loadIdentity
      (_,Size w h) <- get viewport
      ortho2D 0 (fromIntegral w) 0 (fromIntegral h)
      a
    matrixMode $= Modelview 0
    return r
    
-- set up 2D project matrix in -1,-1 .. 1,1
with2DOneOne a = 
  preservingMatrix$do
    loadIdentity
    matrixMode $= Projection
    r <- preservingMatrix$do 
      loadIdentity
      ortho2D (-1) 1 (-1) 1
      a
    matrixMode $= Modelview 0
    return r    

-- drawing rectangles and rounded rectangles with textures

renderTexRect tex x1 y1 x2 y2 z =
  with
  [texture Texture2D        := Enabled,
   textureBinding Texture2D := Just tex] .
  renderPrimitive Quads .
    forM_ [(0,1,x1,y1),(0,0,x1,y2),(1,0,x2,y2),(1,1,x2,y1)] $ \(s,t,x,y) -> do
      texCoord $ TexCoord2 s (t::GLint)
      vertex $ Vertex3 x y (z::GLdouble)
  
renderTexRectAtO tex x y z = renderTexRect tex (-x) (-y) x y z

renderTexRoundedRect tex x1 y1 x2 y2 z r = do
  renderTexRect tex x1 y1 x2 y2 z
  renderRRBorder x1 y1 x2 y2 z r
  
renderTexRoundedRectAtO tex x y z r = renderTexRoundedRect tex (-x) (-y) x y z r

renderRoundedRectAtO x y z r = renderRoundedRect (-x) (-y) x y z r

renderRoundedRect x1 y1 x2 y2 z r = do
  renderPrimitive Quads .
    forM_ [(x1,y1),(x1,y2),(x2,y2),(x2,y1)] $ \(x,y) ->
    vertex $ Vertex3 x y z
  renderRRBorder x1 y1 x2 y2 z r
  
renderRRBorder x1 y1 x2 y2 z r = do
  renderPrimitive Quads .
    forM_ [(x1,y1),(x1-r,y1),(x1-r,y2),(x1,y2),
           (x2,y1),(x2+r,y1),(x2+r,y2),(x2,y2),
           (x1,y1),(x1,y1-r),(x2,y1-r),(x2,y1),
           (x1,y2),(x1,y2+r),(x2,y2+r),(x2,y2)] $ \(x,y) ->
      vertex $ Vertex3 x y z
  renderQuarterCircle x1 y1 (-r) (-r) z
  renderQuarterCircle x2 y1 ( r) (-r) z
  renderQuarterCircle x2 y2 ( r) ( r) z
  renderQuarterCircle x1 y2 (-r) ( r) z

renderQuarterCircle x y rx ry z =
  renderPrimitive TriangleFan .
    forM_ ([(x,y),(x+rx,y)] ++ 
           map (\th -> (x + rx * cos th , y + ry * sin th)) [pi/6,2*pi/6] ++
           [(x,y+ry)]) $ \(x,y) -> vertex $ Vertex3 x y (z::GLdouble)
  

renderCoin tex r d =
  let circf r n    = let th = fromIntegral n * 2 * pi / fromIntegral pts 
                     in (r * cos th, r * sin th)
      inner        = circf (r*0.9)
      outer        = circf r
      pts          = 36
      v x y z      = do texCoord $ TexCoord2 (x/(2*r) + 0.5) (y/(-2*r) + 0.5)
                        vertex $ Vertex3 x y (z::GLdouble)
      vcenter z sn = do normal $ Normal3 0 0 (sn*1::GLdouble)
                        v 0 0 z
      vinner n z sn= do normal $ Normal3 0 0 (sn*1::GLdouble)
                        let (x,y) = inner n
                        v x y z
      vouter n z sn= do let (x,y) = outer n
                        normal $ Normal3 x y 0
                        v x y z
  in 
   with [texture Texture2D        := Enabled,
         textureFunction          := Modulate,
         textureBinding Texture2D := Just tex] $ do
     renderPrimitive TriangleFan $ do
       vcenter (d*2) 1
       forM_ [0..pts] (\n -> vinner n (d*2) 1)
     renderPrimitive QuadStrip .
       forM_ [0..pts] $ \n -> do
         vinner n (d*2) 1
         vouter n d 1
     renderPrimitive QuadStrip .
       forM_ [0..pts] $ \n -> do
         vinner n 0 (-1)
         vouter n d (-1)
     renderPrimitive TriangleFan $ do
       vcenter 0 (-1)
       forM_ [0..pts] (\n -> vinner n 0 (-1))
        
xprod (x1,y1,z1) (x2,y2,z2) =
  Normal3 (y1*z2 - y2*z1) (z1*x2-x1*z2) (x1*y2-y1*x2)
  
toRad i = (realToFrac i)*2*pi/360.0

fromRad :: GLdouble -> GLdouble
fromRad r = r/(2*pi/360.0)

-- mechanism for rendering a texture only when the parameters change

-- data CachedTexture s = CachedTexture (s -> IO (TextureObject,GLdouble))
--                                       (Maybe (s,TextureObject,GLdouble))
type CachedTexture s = CachedResource s (TextureObject,GLdouble)

mkCachedTexture mktex = mkCachedResource mktex (\(t,_) -> deleteObjectNames [t])

withCachedTexture :: (MonadIO m, Ord s,Show s) =>
                     MonadLens m (CachedTexture s) -> s ->
                     (TextureObject -> GLdouble -> m ()) ->
                     m ()
withCachedTexture ctl s act = withCachedResource ctl s (uncurry act)
-- withCachedTexture ctl s act = do
--   CachedTexture mktex cache <- mlGet ctl
--   let render = do (tex,ar) <- liftIO $ mktex s
--                   mlSet ctl (CachedTexture mktex (Just (s,tex,ar)))
--                   return (tex,ar)
--   (tex,ar) <- case cache of 
--     Just (s',tex,ar) | s == s' -> return (tex,ar)
--                      | otherwise -> do liftIO $ deleteObjectNames [tex]
--                                        render
--     _ -> render
--   act tex ar

-- mechanism for acquiring resources / doing expensive computation 
-- only when the parameters change

data CachedResource s r = 
  CachedResource { crSetup :: s -> IO r,
                   crTeardown :: r -> IO (),
                   crCache :: Map s r }

mkCachedResource setup teardown = CachedResource setup teardown M.empty

withCachedResource :: (MonadIO m, Ord s, Show s) =>
                     MonadLens m (CachedResource s r) -> s ->
                     (r -> m ()) -> m ()
withCachedResource ctl s act = do
  CachedResource setup teardown cache <- mlGet ctl
  let render = do r <- liftIO $ setup s
                  liftIO.putStrLn $ "Rendering for parms " ++ show s ++
                    " cache size is now " ++ show (M.size cache)
                  mlSet ctl (CachedResource setup teardown (M.insert s r cache))
                  return r
  r <- case M.lookup s cache of 
    Just r -> return r
    Nothing -> render
  act r

-- A mechanism for building up a Data.Map of cached resources against
-- different parameters

