--
-- force-based 3D graph layout.
--

import qualified Data.Set as S
import Data.Set(Set)
import qualified Data.Map as M
import Data.Map(Map,(!))
import qualified Data.Foldable as F

import Data.IORef
import Data.Monoid
import Control.Applicative

import Graphics.Rendering.OpenGL hiding (Sink)
import Graphics.UI.GLUT hiding (Sink)

import Deus.Reactive
import Deus.GLUtils

-- represent a graph as a set of nodes and a set of edges

data Graph a = 
  Graph { gNodes :: Set a,
          gEdges :: Set (a,a) } deriving Show

-- ideal length of springs

ideal :: GLdouble
ideal = 1

-- a little bit of vector math in R3 will be useful
Vector3 x0 y0 z0 <+> Vector3 x1 y1 z1 = Vector3 (x0+x1) (y0+y1) (z0+z1)
Vector3 x0 y0 z0 <-> Vector3 x1 y1 z1 = Vector3 (x0-x1) (y0-y1) (z0-z1)
sc *^ Vector3 x0 y0 z0 = Vector3 (sc*x0) (sc*y0) (sc*z0)

vsum :: Num t => [Vector3 t] -> Vector3 t
vsum = foldr (<+>) (Vector3 0 0 0)

mag2 (Vector3 x y z) = x*x + y*y + z*z
mag v = sqrt (mag2 v)


-- a simulation is a map from nodes to positions + velocities

data Simulation a = 
  Simulation {
    sPositions :: Map a (Vector3 GLdouble),
    sVelocities :: Map a (Vector3 GLdouble)
    } deriving Show

-- initial simulation has all nodes with zero velocity
-- we want to choose arbitrary starting positions close to the origin
-- and with low probability of being coplanar. 
positions = map (\n -> (Vector3 (sin n) (cos n) (sin (n*5.17)))) [1..]

initialSimulation :: Ord a => Graph a -> Simulation a
initialSimulation g = 
  Simulation 
  (M.fromList (zip (S.elems . gNodes $ g) positions))
  (M.fromList (zip (S.elems . gNodes $ g) (repeat (Vector3 0 0 0))))

-- step simulation 
stepSimulation :: Ord a => 
                  Graph a ->
                  GLdouble ->
                  Simulation a ->
                  Simulation a
stepSimulation g delta_t sim =
  let
    -- move all positions by current velocities
    newPositions = 
      M.mapWithKey 
      (\k p -> p <+> (delta_t *^ (sVelocities sim ! k)))
      (sPositions sim)
    -- move all velocities by acceleration
    newVelocities = 
      M.mapWithKey 
      (\k p -> p <+> (delta_t *^ (acceleration k)))
      (sVelocities sim)
    -- calculate acceleration at each node
    acceleration k =
      let 
        -- position of this node
        kPos = sPositions sim ! k
        -- displacement vector to other node
        disp k' = (sPositions sim ! k') <-> kPos
        -- unit displacement vector
        unit k' = let d = disp k' in (1/mag d) *^ d
      in 
       vsum (
         -- r2 repulsion from each other node
         map (\k' -> (-1 / mag2 (disp k')) *^ unit k')
         (F.toList $ S.delete k (gNodes g))
         ++
         -- linear attraction from attached node
         (map (\k' -> (mag (disp k') - ideal) *^ unit k') .
          concatMap (\(a,b) -> 
                      if a == k then [b] else if b == k then [a] else []) .
          F.toList $ gEdges g)
         ++
         -- damping term
         [(-0.5) *^ (sVelocities sim ! k)]
         )
  in Simulation newPositions newVelocities

-- example data

mkGraph edges = Graph { gNodes = S.fromList (map fst edges ++ map snd edges),
                        gEdges = S.fromList edges }                

black,white,red,green,blue,yellow :: Color3 GLfloat
black  = Color3 0.2 0.2 0.2
white  = Color3 1 1 1 
red    = Color3 1 0 0 
green  = Color3 0 1 0 
blue   = Color3 0 0 1
yellow = Color3 1 1 0


-- platonic solids

allred = const (red,0.1)

cube = 
  (mkGraph [('A','B'),('B','C'),('C','D'),('D','A'),
            ('E','F'),('F','G'),('G','H'),('H','E'),
            ('A','E'),('B','F'),('C','G'),('D','H')],
   allred)

tetrahedron = 
  (mkGraph [('A','B'),('A','C'),('A','D'),
            ('B','C'),('C','D'),('D','B')],
   allred)

octahedron = 
  (mkGraph [('A','B'),('A','C'),('A','D'),('A','E'),
            ('B','C'),('C','D'),('D','E'),('E','B'),
            ('F','B'),('F','C'),('F','D'),('F','E')],
   allred)

-- can picture dodecahedron as a north pole (pentagon), south pole, and 
-- equator (10 points on equator)

dodecahedron = 
  (mkGraph [("N1","N2"),("N2","N3"),("N3","N4"),("N4","N5"),("N5","N1"),
            ("S1","S2"),("S2","S3"),("S3","S4"),("S4","S5"),("S5","S1"),
            ("E1","E2"),("E2","E3"),("E3","E4"),("E4","E5"),("E5","E6"),
            ("E6","E7"),("E7","E8"),("E8","E9"),("E9","E0"),("E0","E1"),
            ("N1","E1"),("N2","E3"),("N3","E5"),("N4","E7"),("N5","E9"),
            ("S1","E2"),("S2","E4"),("S3","E6"),("S4","E8"),("S5","E0")
           ],
   (\xs -> case head xs of 'N' -> (white,0.1)
                           'S' -> (black,0.1)
                           'E' -> (red,0.1)))

-- some schemes for solids

prism n = 
  (mkGraph ([ ((p,m),(p,(m `mod` n) + 1)) | p <- "NS", m <- [1..n] ] ++
            [ (('N',m),('S',m)) | m <- [1..n] ]),
   (\(c,_) -> case c of 'N' -> (green,0.1)
                        'S' -> (red,0.1)))
  
trapezohedron n =
  (mkGraph ([ (('N',1),('E',m*2)) | m <- [1..n] ] ++
           [ (('S',1),('E',m*2 - 1)) | m <- [1..n] ] ++
           [ (('E',m),('E',m `mod` (n*2) + 1)) | m <- [1..n*2] ]
          ),
   (\(c,_) -> case c of 'N' -> (green,0.1)
                        'E' -> (white,0.1)
                        'S' -> (red,0.1)))

-- chemicals

atoms xs = case head xs of 'C' -> (red,0.2)
                           'H' -> (blue,0.1)
                           
ch4 = 
  (mkGraph [ ("C",h) | h <- ["H1","H2","H3","H4"] ],
   atoms)

c2h6 = (mkGraph (("C1","C2") : 
                 [ ("C1",h) | h <- ["H1","H2","H3"] ] ++ 
                 [ ("C2",h) | h <- ["H4","H5","H6"] ] ),
        atoms)
  
c3h8 = (mkGraph (("C1","C2") : ("C2","C3") : 
                [ ("C1",h) | h <- ["H1","H2","H3"] ] ++ 
                [ ("C2",h) | h <- ["H4","H5"] ] ++
                [ ("C3",h) | h <- ["H6","H7","H8"] ] ),
        atoms)

-- well known graph-theoretical constructs

complete n = (mkGraph [(m,p) | m <- [1..n], p <- [m+1..n]],allred)
bipartite n m = 
  (mkGraph [(Left p,Right q) | p <- [1..n], q <- [1..m]],
   (either (\_ -> (red,0.1)) (\_ -> (green,0.1))))

-- haskell data graphs

l123 = (mkGraph [(":1","1"),(":1",":2"),
                 (":2","2"),(":2",":3"),
                 (":3","3"),(":3","[]")],
        (\xs -> case head xs of ':' -> (white,0.2)
                                '[' -> (white,0.1)
                                _   -> (red,0.1)))


-- openGL code to visualise

glinit = do
  (progName,args) <-  getArgsAndInitialize
--  initialDisplayMode $= [ DoubleBuffered, RGBAMode, WithDepthBuffer, Multisampling] 
  initialDisplayCapabilities $=
    [With DisplayDouble,
     With DisplayRGBA,
     With DisplayDepth,
     Where DisplaySamples IsAtLeast 2]
  createWindow progName
  multisample $= Enabled
  
  windowSize         $= (Size 800 600)
  depthFunc          $= Just Less
  
  shadeModel         $= Smooth

  normalize          $= Enabled
  autoNormal         $= Enabled
  lighting           $= Enabled
  position (Light 0) $= Vertex4 2 2 (2)  1
  ambient (Light 0)  $= Color4 0.0 0.0 0.0 1
  diffuse (Light 0)  $= Color4 0.8 0.8 0.8 1
  specular (Light 0) $= Color4 1 1 1 1
  light (Light 0)    $= Enabled

  blend              $= Enabled
  blendFunc          $= (SrcAlpha,OneMinusSrcAlpha)
  
main = do
  glinit
  
  let (graph,cols) = c3h8
      
  -- how often to evaluate the simulation, as a fraction of a second
  let ticklength = 0.01
  simChangesE <- timewiseIterate (round (ticklength * 1e6))
               (stepSimulation graph 0.02) 
               (initialSimulation graph)
  let sim = initialSimulation graph `Stepper` (simChangesE)
  
  eref <- newIORef (mempty)
  (kpSink,keypresses) <- makeSinkEvent
  keyboardMouseCallback $= Just (reactiveKeyboardMouseCallback kpSink eref)  
  
  let theta = 0 `accumR` (fmap (\k -> case k of 
                                   SpecialKey KeyLeft -> (+0.1)
                                   SpecialKey KeyRight -> (subtract 0.1)
                                   _ -> id) keypresses)
  let distance = 5 `accumR` (fmap (\k -> case k of 
                                   SpecialKey KeyPageUp -> (*1.05)
                                   SpecialKey KeyPageDown -> (/1.05)
                                   _ -> id) keypresses)
              
  bref <- newIORef ((,,) <$> rToB sim <*> rToB theta <*> rToB distance)

  displayCallback $= 
    do (sim,th,rad) <- sampleBRef bref
       draw graph cols sim th rad

  reshapeCallback $= Just reshape
  addTimerCallback 0 (timer eref)
  
  postRedisplay Nothing
  
  mainLoop

timer eref = 
  do postRedisplay Nothing
     runReactorRef eref
--     System.Mem.performGC
     -- cap framerate crudely to 60fps
     addTimerCallback 16 (timer eref)

-- adapter to populate a Chan with keys
reactiveKeyboardMouseCallback :: Sink Key -> IORef (Event (IO ())) ->
                  Key -> KeyState -> Modifiers -> Position -> IO ()
reactiveKeyboardMouseCallback chan eref k Down m p = do
  sendEvent chan k
  runReactorRef eref
reactiveKeyboardMouseCallback _    eref _ _    _ _ = 
  runReactorRef eref

draw graph cols sim theta camdistance = do
  loadIdentity
  hint PerspectiveCorrection     $= Nicest
  depthMask $= Enabled
  depthFunc $= Just Lequal
  
  colorMaterial $= Just (FrontAndBack,AmbientAndDiffuse)
  clear [ColorBuffer,DepthBuffer]

  lookAt (Vertex3 (camdistance*sin theta) 0 (camdistance*cos theta)) 
    (Vertex3 0 0 0) (Vector3 0 1 0)
  
  -- automatically scale down large figures
  let sc = 1 / (maximum . map mag . M.elems . sPositions $ sim)
  scale sc sc sc
  
  color $ Color3 1 0 (0::GLfloat)
  sequence_ . M.elems $ 
    M.mapWithKey 
    (\n p ->
        preservingMatrix $ do
          let (col,radius) = cols n
          color col
          translate p
          renderObject Solid (Sphere' radius 12 12)
    )
    (sPositions sim)
  
  color $ Color3 1 1 (0::GLfloat)
  F.forM_
    (gEdges graph)
    (\(n,m) -> 
      renderPrimitive Lines $ do
        vertex . vc2vx $ sPositions sim ! n
        vertex . vc2vx $ sPositions sim ! m
    )

  swapBuffers
  
vc2vx (Vector3 x y z) = Vertex3 x y z